% [out, out_high, out_low, high_weight , low_weight, sigma] = ALF7(im1, im2)
% SSIM-based linear adaptive image fusion
% Inputs:
%	im1: first image
%	im2: second image
%	L: operating windows, default (5x5)
%
% Usage:
%	[out] = ALF_SigmaEstimation(im1, im2, 3)
%	
% 
%	NTC June 2012
%	NTC Sept 2012
%
function [out, out_high, out_low, high_weight, low_weight, lp_sigma] = ALF_SigmaEstimation(im1, im2, varargin)

	if(nargin < 2)
		error('Not enough input argument: ALF(im1, im2)');
	elseif(nargin == 2)
		L = 2;
	else
		L = varargin{1};
	end

	if(size(im1,1) ~= size(im2,1) || size(im1,2) ~= size(im2,2))
		error('Size mismatch: a and b have different dimension.');
	end

	tstart = tic;
	window = (1/(2*L+1))^2 * ones(2*L+1,2*L+1);
	mean_a = filter2(window, im1, 'same');
	mean_b = filter2(window, im2, 'same');
	sigma_a_square = filter2(window, im1.*im1, 'same') - mean_a.^2;
	sigma_b_square = filter2(window, im2.*im2, 'same') - mean_b.^2;


	% temp variables for evaluating Q
	[M,N] = size(im1);
	if( mod(M,2) == 0)
		half_row = M/2;
	else
		half_row = (M-1)/2;
	end

	if( mod(N,2) == 0)
		half_col = N/2;
	else
		half_col = (N-1)/2;
	end

	[xx, yy] = meshgrid(-half_col:1:(N-half_col-1), -half_row:1:(M-half_row-1));
	im1_fft = fft2(im1);
	im2_fft = fft2(im2);
	diff_fft = fft2(im1-im2);
	sum_fft = fft2(im1+im2);
	mean_a_fft = fft2(mean_a);
	mean_b_fft = fft2(mean_b);
	sigma_a_square_fft = fft2(sigma_a_square);
	sigma_b_square_fft = fft2(sigma_b_square);



	% find the optimized sigma of the lowpass filter
	min_sigma = .1;
	max_sigma = sqrt( (M^2 + N^2) / 8 );

	sigma_list = [min_sigma:0.3:max_sigma];
	y = zeros(1,length(sigma_list));
	for i=2:length(sigma_list)
		sigma = sigma_list(i);
		H = exp( - (xx.^2 + yy.^2) / (2*sigma^2) );

		%showimage(log10(1+abs(H)));
		t1 = real(ifft2( mean_a_fft .* ifftshift(H)));
		t2 = real(ifft2( (mean_a_fft + mean_b_fft) .* ifftshift(H)));
		%t1 = real(ifft2( im1_fft .* ifftshift(H)));
		%t2 = real(ifft2( (im1_fft + im2_fft) .* ifftshift(H)));
		t3 = real(ifft2( (diff_fft) .* ifftshift(H)));

		t4 = real(ifft2( sigma_a_square_fft .* ifftshift(H)));
		t5 = real(ifft2( (sigma_a_square_fft + sigma_b_square_fft) .* ifftshift(H)));

		low_weight = medfilt2(t1 ./ (t2 + eps), [3 3]);
		high_weight = medfilt2(t4 ./ (t5 + eps), [3 3]);
		out = low_weight .* t3 + high_weight .* ((im1-im2) - t3) + im2;
		y(i) = metricPeilla(im1,im2,FullScaleStretch(out),3);
		if(y(i) < y(i-1))
			lp_sigma = sigma_list(i-1);
			break;
		end
	end


	H = exp( - (xx.^2 + yy.^2) / (2*lp_sigma^2) );
	t1 = real(ifft2( mean_a_fft .* ifftshift(H)));
	t2 = real(ifft2( (mean_a_fft + mean_b_fft) .* ifftshift(H)));
	t3 = real(ifft2( (diff_fft) .* ifftshift(H)));
	t4 = real(ifft2( sigma_a_square_fft .* ifftshift(H)));
	t5 = real(ifft2( (sigma_a_square_fft + sigma_b_square_fft) .* ifftshift(H)));
	
	low_weight = medfilt2(t1 ./ (t2 + eps), [3 3]);
	high_weight = medfilt2(t4 ./ (t5 + eps), [3 3]);

	out = low_weight .* t3 + high_weight .* ((im1-im2) - t3) + im2;
	out_low = low_weight .*  real(ifft2(im1_fft .* ifftshift(H))) + (1-low_weight) .* real(ifft2( im2_fft .* ifftshift(H)));
	out_high = high_weight .*  (im1 - real(ifft2(im1_fft .* ifftshift(H)))) + ...
			   (1-high_weight) .* (im2 - real(ifft2( im2_fft .* ifftshift(H))));

	telapsed = toc(tstart)
end
